{ 
    open Parser 
    
    
    (*Increment the line number tracked by lexbuf*)
    let incrLineNum lexbuf =
        let pos = lexbuf.Lexing.lex_curr_p in
            lexbuf.Lexing.lex_curr_p <- { pos with
                Lexing.pos_lnum = pos.Lexing.pos_lnum + 1;
                Lexing.pos_bol = pos.Lexing.pos_cnum;
            }
            
}


let floatLit = ['0'-'9']+['.']['0'-'9']* (['e''E'](['-''+']?['0'-'9']+))?
| ['0'-'9']*['.']['0'-'9']+ (['e''E'](['-''+']?['0'-'9']+))?
| ['0'-'9']*['.']?['0'-'9']+ (['e''E']['-''+']?['0'-'9']+)



rule token = parse
  [' ' '\t']     { token lexbuf }
| ['\r' '\n']    { incrLineNum lexbuf; token lexbuf }
| '%'            { comment lexbuf }
| '"'            { stringParse "" lexbuf}
| '('            { LPAREN }
| ')'            { RPAREN }
| '['            { LBRACKET }
| ']'            { RBRACKET }
| ';'            { SEMI }
| ','            { COMMA }
| ':'            { COLON }
| '^'            { POWER }  
| '*'            { MUL }  
| ".*"           { EMUL}
| "./"           { ERDIV}  (*element by element right divid*)
| '\\'            { LDIV}   (*matrix left divide*)
| ".\\"           {ELDIV}  (*element by element left divide*)
| '/'            { DIVIDE } (*divide or matrix right divide*)
| '\''           {TRANSPOSE}
| '+'            { PLUS }
| '-'            { MINUS }
| '='            { ASSIGN }
| "=="           { EQ }
| "!="           { NOTEQ }
| '<'            { LT }
| "<="           { LTEQ }
| ">"            { GT }
| ">="           { GTEQ }
| "&"            { AND }
| "|"            { OR }
| "!"            { NOT }
| "if"           { IF }
| "else"         { ELSE }
| "elseif"       { ELSEIF }
| "true"  as lxm { BOOL(lxm) }
| "false" as lxm { BOOL(lxm) }
| "for"          { FOR }
| "while"        { WHILE }
| "continue"     { CONTINUE }
| "break"        { BREAK }
| "end"          { END }
| "function"     { FUNCTION }
| "boolean"  	{ DATATYPE("boolean") }
| "int"  	{ DATATYPE("int") }
| "float"  	{ DATATYPE("float") }
| "string" 	{ DATATYPE("string") }
| "matrix" 	{ DATATYPE("matrix") }
| "newmat"      { NEWMAT }
| ['0'-'9']+ as lxm  { LITERAL(int_of_string lxm) }
| floatLit as lxm    { FLOATLIT(float_of_string lxm) }
| ['a'-'z' 'A'-'Z']['a'-'z' 'A'-'Z' '0'-'9' '_']* as lxm { ID(lxm) }
| eof { EOF }
| _ as char { 
            let pos = lexbuf.Lexing.lex_curr_p in
            raise (Failure("Illegal character: " ^ Char.escaped char
            ^ " in line #" ^ (string_of_int pos.Lexing.pos_lnum))) }

and comment = parse
  "\n" { incrLineNum lexbuf; token lexbuf }
| _    { comment lexbuf }

and stringParse str = parse
| '"'  { STRING(str) } (* end of a string literal *)
| [^ '"'] as c  { stringParse (str ^ String.make 1 c) lexbuf } (*everything except quote is counted as a character inside the string*)
|  "\n" {let currPos = lexbuf.Lexing.lex_curr_p in
            raise (Failure("Unclosed string literal, in line #" ^ (string_of_int currPos.Lexing.pos_lnum))) }
|  eof  {let currPos = lexbuf.Lexing.lex_curr_p in
            raise (Failure("Unclosed string literal, in line #" ^ (string_of_int currPos.Lexing.pos_lnum))) }