open Ast

module NameMap = Map.Make(struct
  type t = string
  let compare x y = Pervasives.compare x y
end)

let loop_flag = ref false;;
let string_of_program (vars, funcs) =
  (* Put function declarations in a symbol table *)
  let func_decls = List.fold_left
        (fun funcs fdecl -> 
            if NameMap.mem fdecl.fname funcs then
                raise (Failure ("Function: '" ^ fdecl.fname ^ 
                    "' has already been defined."))
            else            
                NameMap.add fdecl.fname fdecl funcs) NameMap.empty funcs
    in
    let blackhole e = () in
    
    let rec string_of_fdecl globals fdecl actuals =
    
    let rec string_of_expr globals locals fname = function
	 Literal(l) -> { varname = string_of_int l;vartype = "int" }
      | Floatlit(f) -> { varname = string_of_float f;vartype = "float" }
      | String(s) -> { varname = s;vartype = "string" }
      | Bool(s) -> { varname = s;vartype = "boolean" }      
      | Id(s) ->
            if NameMap.mem s locals then 
    	        NameMap.find s locals
            else if NameMap.mem s globals then 
    	        NameMap.find s globals
    	    else raise (Failure ("Undeclared identifier found in expression: '" 
    	        ^ s ^ "' in function: '" ^ fname ^ "'"))
      | Datatype(t) -> 
            if t = "parseerror" then
                raise ((Failure ("")))
            else { varname = "null" ;vartype = t }
      | Paren (e) -> string_of_expr globals locals fname e 
      | Transpose(e)->let v = string_of_expr globals locals fname e in 
            if v.vartype = "matrix" then
                v
            else
                raise (Failure ("transpose operator ' can only be applied to matrix expressions" 
                                ^ ", but here used with '" ^ v.varname ^ "', type: " 
                                ^ v.vartype ^ " in function: '" ^ fname ^ "'"))           
      | Uminus(e) -> let v = string_of_expr globals locals fname e in 
            if v.vartype = "int" || v.vartype = "float" then
                v
            else
                raise (Failure ("unary - operator can only be applied to int and float expressions" 
                                ^ ", but here used with '" ^ v.varname ^ "', type: " 
                                ^ v.vartype ^ " in function: '" ^ fname ^ "'"))
      | Not(e) -> let v = string_of_expr globals locals fname e in
            if v.vartype = "boolean" then
                v
            else
                raise (Failure ("not operator can only be applied to boolean expression" 
                                ^ ", but here used with '" ^ v.varname ^ "', type: " 
                                ^ v.vartype ^ " in function: '" ^ fname ^ "'"))
      | Binop(e1, o, e2) ->
          let v1 = string_of_expr globals locals fname e1 in
          (match o with
             Add ->
             let v2 = string_of_expr globals locals fname e2 in
               if v1.vartype = "string" then
                 if v2.vartype = "matrix" then
                   raise (Failure ("Cannot concatenate " ^ v2.vartype ^ " type with string type " ^
                                   "in function: '" ^ fname ^ "'"))
                 else v1
               else if v2.vartype = "string" then
                 if v1.vartype = "matrix" then
                   raise (Failure ("Cannot concatenate " ^ v1.vartype ^ " type with string type " ^
                                   "in function: '" ^ fname ^ "'"))
                 else v2
               else if v1.vartype = "matrix" && v2.vartype = "int" then
                      v1
               else if v1.vartype = "matrix" && v2.vartype = "float" then
                      v1
               else if v1.vartype = "int" && v2.vartype = "matrix" then
                      v2
               else if v1.vartype = "float" && v2.vartype = "matrix" then
                      v2
               else if v1.vartype = "matrix" && v2.vartype = "matrix" then
                      v1
               else if v1.vartype = "float" && v2.vartype="int" then
                    v1
               else if v1.vartype = "int" && v2.vartype="float" then
                    v2
               else if v1.vartype = "boolean" || v2.vartype = "boolean" then
                 raise (Failure ("+ operator cannot be applied to boolean operands" 
                                  ^ ", but here used with '" ^ v1.varname ^ "', type: " 
                                  ^ v1.vartype ^ " and '" ^ v2.varname ^ "', type: " 
                                  ^ v2.vartype ^ " in function: '" ^ fname ^ "'"))
               else if v1.vartype <> v2.vartype then
                 raise (Failure ("Type mismatch in binary operation between: '" ^ 
                                  v1.varname ^ "', type: " ^ v1.vartype ^ " and '" ^ v2.varname
                                  ^ "', type: " ^ v2.vartype ^ " in function: '" ^ fname 
                                  ^ "'"))
               else v1 
    	   | Sub -> 
             let v2 = string_of_expr globals locals fname e2 in
                if v1.vartype = "string" || v2.vartype = "string" then
                  raise (Failure ("Operators '-' " ^
                                  "cannot be applied to a string, in function: '" ^ fname ^ "'"))
               else if v1.vartype = "matrix" && v2.vartype = "int" then
                      v1
               else if v1.vartype = "matrix" && v2.vartype = "float" then
                      v1
               else if v1.vartype = "int" && v2.vartype = "matrix" then
                      v2
               else if v1.vartype = "float" && v2.vartype = "matrix" then
                      v2
               else if v1.vartype = "matrix" && v2.vartype = "matrix" then
                      v1
               else if v1.vartype = "float" && v2.vartype="int" then
                    v1
               else if v1.vartype = "int" && v2.vartype="float" then
                    v2               
               else if v1.vartype = "boolean" || v2.vartype = "boolean" then
                 raise (Failure ("- operator cannot be applied to boolean operands" 
                                  ^ ", but here used with '" ^ v1.varname ^ "', type: " 
                                  ^ v1.vartype ^ " and '" ^ v2.varname ^ "', type: " 
                                  ^ v2.vartype ^ " in function: '" ^ fname ^ "'"))
                else if v1.vartype <> v2.vartype then 
                  raise (Failure ("Type mismatch in binary operation between: '" ^ 
                                   v1.varname ^ "', type: " ^ v1.vartype ^ " and '" ^ v2.varname
                                   ^ "', type: " ^ v2.vartype ^ " in function: '" ^ fname 
                                   ^ "'")) 
                else v1
           | Equal | Neq ->
             let v2 = string_of_expr globals locals fname e2 in
                if v1.vartype <> v2.vartype then 
                  raise (Failure ("Type mismatch in binary operation between: '" ^ 
                                   v1.varname ^ "', type: " ^ v1.vartype ^ " and '" ^ v2.varname
                                   ^ "', type: " ^ v2.vartype ^ " in function: '" ^ fname 
                                   ^ "'")) 
                else { varname = "null"; vartype = "boolean" }
    	   | Mul -> 
             let v2 = string_of_expr globals locals fname e2 in
                if v1.vartype = "string" || v2.vartype = "string" then
                  raise (Failure ("Operators '*', '/' " ^
                                  "cannot be applied to a string, in function: '" ^ fname ^ "'"))
               else if v1.vartype = "matrix" && v2.vartype = "int" then
                      v1
               else if v1.vartype = "matrix" && v2.vartype = "float" then
                      v1
               else if v1.vartype = "int" && v2.vartype = "matrix" then
                      v2
               else if v1.vartype = "float" && v2.vartype = "matrix" then
                      v2
               else if v1.vartype = "matrix" && v2.vartype = "matrix" then
                      v1
               else if v1.vartype = "float" && v2.vartype="int" then
                    v1
               else if v1.vartype = "int" && v2.vartype="float" then
                    v2               
               else if v1.vartype = "boolean" || v2.vartype = "boolean" then
                 raise (Failure ("* operators cannot be applied to boolean operands" 
                                  ^ ", but here used with '" ^ v1.varname ^ "', type: " 
                                  ^ v1.vartype ^ " and '" ^ v2.varname ^ "', type: " 
                                  ^ v2.vartype ^ " in function: '" ^ fname ^ "'"))
                else if v1.vartype <> v2.vartype then 
                  raise (Failure ("Type mismatch in binary operation between: '" ^ 
                                   v1.varname ^ "', type: " ^ v1.vartype ^ " and '" ^ v2.varname
                                   ^ "', type: " ^ v2.vartype ^ " in function: '" ^ fname 
                                   ^ "'")) 
                else v1
           | Div -> 
	                let v2 = string_of_expr globals locals fname e2 in
	                   if v1.vartype = "string" || v2.vartype = "string" then
	                     raise (Failure ("Operators '*', '/' " ^
	                                     "cannot be applied to a string, in function: '" ^ fname ^ "'"))
	                  else if v1.vartype = "matrix" && v2.vartype = "int" then
	                         v1
	                  else if v1.vartype = "matrix" && v2.vartype = "float" then
	                         v1
	                  else if v1.vartype = "int" && v2.vartype = "matrix" then
	                         v2
	                  else if v1.vartype = "float" && v2.vartype = "matrix" then
	               		v2
	               else if v1.vartype = "float" && v2.vartype="int" then
	                    v1
	               else if v1.vartype = "int" && v2.vartype="float" then
	                    v2	                         

	                  else if v1.vartype = "matrix" && v2.vartype = "matrix" then
                      		 raise(Failure("/ operator cannot be applied to matrix operands"
                      		 ^ ", but here used with '" ^ v1.varname ^ "', type: " 
				 ^ v1.vartype ^ " and '" ^ v2.varname ^ "', type: " 
	                         ^ v2.vartype ^ " in function: '" ^ fname ^ "'"))
	                   else if v1.vartype = "boolean" || v2.vartype = "boolean" then
	                    raise (Failure ("/ operator cannot be applied to boolean operands" 
	                                     ^ ", but here used with '" ^ v1.varname ^ "', type: " 
	                                     ^ v1.vartype ^ " and '" ^ v2.varname ^ "', type: " 
	                                     ^ v2.vartype ^ " in function: '" ^ fname ^ "'"))
	                   else if v1.vartype <> v2.vartype then 
	                     raise (Failure ("Type mismatch in binary operation between: '" ^ 
	                                      v1.varname ^ "', type: " ^ v1.vartype ^ " and '" ^ v2.varname
	                                      ^ "', type: " ^ v2.vartype ^ " in function: '" ^ fname 
	                                      ^ "'")) 
	                  else if v1.vartype ="matrix" && v2.vartype="matrix" then
	                    raise (Failure ("/ operator cannot be applied to matrix operands" 
	                                     ^ ", but here used with '" ^ v1.varname ^ "', type: " 
	                                     ^ v1.vartype ^ " and '" ^ v2.varname ^ "', type: " 
	                                     ^ v2.vartype ^ " in function: '" ^ fname ^ "'"))
	                  
                	else v1
           | Emul | Erdiv | Ldiv|Eldiv->
		let v2 = string_of_expr globals locals fname e2 in
                if v1.vartype = "matrix" && v2.vartype = "matrix" then
                    v1
                else
                  raise (Failure ("The operands for operators '.*', './', and '\\' " ^
                                  "must be matrices, in function: '" ^ fname ^ "'"))   
    	   | Less | Leq | Greater | Geq -> 
             let v2 = string_of_expr globals locals fname e2 in
                if v1.vartype = "string" || v2.vartype = "string" then
                  raise (Failure ("Operators '<', '<=', '>', '>=' " ^
                                  "cannot be applied to a string, in function: '" ^ fname ^ "'"))
               else if v1.vartype = "boolean" || v2.vartype = "boolean" then
                 raise (Failure ("<,<=,>,>= operators cannot be applied to boolean operands" 
                                  ^ ", but here used with '" ^ v1.varname ^ "', type: " 
                                  ^ v1.vartype ^ " and '" ^ v2.varname ^ "', type: " 
                                  ^ v2.vartype ^ " in function: '" ^ fname ^ "'"))
		else if v1.vartype = "matrix" || v2.vartype = "matrix" then
                 raise (Failure ("<,<=,>,>= operators cannot be applied to matrix operands" 
                                  ^ ", but here used with '" ^ v1.varname ^ "', type: " 
                                  ^ v1.vartype ^ " and '" ^ v2.varname ^ "', type: " 
                                  ^ v2.vartype ^ " in function: '" ^ fname ^ "'"))                      
                else if v1.vartype <> v2.vartype then 
                  raise (Failure ("Type mismatch in binary operation between: '" ^ 
                                   v1.varname ^ "', type: " ^ v1.vartype ^ " and '" ^ v2.varname
                                   ^ "', type: " ^ v2.vartype ^ " in function: '" ^ fname 
                                   ^ "'")) 
                else { varname = "null"; vartype = "boolean" }
    	   | And | Or -> 
             let v2 = string_of_expr globals locals fname e2 in
                if v1.vartype = "boolean" && v2.vartype = "boolean" then
                    v1
                else
                    raise (Failure ("and, or operators can only be applied to boolean expressions" 
                                  ^ ", but here used with '" ^ v1.varname ^ "', type: " 
                                  ^ v1.vartype ^ " and '" ^ v2.varname ^ "', type: " 
                                  ^ v2.vartype ^ " in function: '" ^ fname ^ "'"))
          | Power ->
            let v2 = string_of_expr globals locals fname e2 in
            if v1.vartype = "int" && v2.vartype = "int" then 
                v1
            else if v1.vartype = "float" && v2.vartype = "float" then 
                v1
            else if v1.vartype = "matrix" && v2.vartype = "int" then 
                v1
            else if v1.vartype = "matrix" && v2.vartype = "float" then 
                v1                
               else if v1.vartype = "float" && v2.vartype="int" then
                    v1
               else if v1.vartype = "int" && v2.vartype="float" then
                    v2
            else
	            raise (Failure ("For ^ operator, if the first operand is int or float, the type of the second argument must be the same as the first.\n"^ 
	            "If the first operand is matrix, the type of the second argument must be either int or float.\n"
	            ^ "but here got '" ^ v1.vartype ^ ": " ^ v1.varname ^ "' and '" 
	            ^ v2.vartype ^ ": " ^ v2.varname ^ "' in function: '" ^ fname ^ "'")))
      | Assign(v, e) -> 
            if NameMap.mem v locals then
                let var1 = NameMap.find v locals in
                let var2 = string_of_expr globals locals fname e in
    	        if var1.vartype <> var2.vartype then 
    	            raise (Failure ("Type mismatch in assignment operation between: '" 
    	                ^ var1.varname ^ "', type: " ^ var1.vartype ^ " and '" 
    	                ^ var2.varname ^ "', type: " ^ var2.vartype 
    	                ^ " in function: '" ^ fname ^ "'")) 
    	        else var1
    	  else if NameMap.mem v globals then
    	        let var1 = NameMap.find v globals in
                let var2 = string_of_expr globals locals fname e in
    	        if var1.vartype <> var2.vartype then 
    	            raise (Failure ("Type mismatch in assignment operation between: '" 
    	                ^ var1.varname ^ "', type: " ^ var1.vartype ^ " and '" 
    	                ^ var2.varname ^ "', type: " ^ var2.vartype 
    	                ^ " in function: '" ^ fname ^ "'")) 
    	        else var1
    	  else raise (Failure ("Undeclared identifier found in assignment: '" 
    	        ^ v ^ "' in function: '" ^ fname ^ "'"))
      | AssignMatrix(id, indexList, value) ->  
                let varvalue = string_of_expr globals locals fname value in  
                if (List.length indexList !=2) then raise (Failure("Matrix assignment for " ^id ^" requires 2 indices"))
                else
                let checkIndices globals locals fname matrix indexList =
			List.iter
			(fun exp ->
				let v = string_of_expr globals locals fname exp in
			      	if v.vartype <> "int" then
			      		raise (Failure ("Invalid index value: all indices "
			      	        ^ " must be int type. matrix: '" ^ matrix ^ "' ;"
			                ^"function: '" ^ fname ^ "'"))) indexList
		in
       
                let chkIn=checkIndices globals locals fname id indexList in
                blackhole chkIn;
                if  varvalue.vartype <> "float" && varvalue.vartype <> "int" then  
                    raise (Failure ("Type mismatch in matrix assignment."   
                        ^ " The value must be of type float or int but here used with type '"
                        ^ varvalue.vartype ^ "' in function: '" ^ fname ^ "'"))  
                else
                    if NameMap.mem id locals then  
                        let varId = NameMap.find id locals in  
                        if varId.vartype <> "matrix" then  
                            raise (Failure ("Expecting matrix type but got '" ^ varId.vartype ^  
                                "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'"))    
                        else { varname = "null";vartype = "int" }   
                    else if NameMap.mem id globals then  
                        let varId = NameMap.find id globals in  
                        if varId.vartype <> "matrix" then  
                        raise (Failure ("Expecting matrix type but got '" ^ varId.vartype ^  
                            "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'"))    
                        else { varname = "null";vartype = "int" }
            	  else raise (Failure ("Undeclared identifier found in assignment: '" 
            	        ^ id ^ "' in function: '" ^ fname ^ "'"))
      | MatrixAccess(id, indexList) ->
              if (List.length indexList !=2) then raise (Failure("Matrix access for " ^id ^" requires 2 indices"))
                  else
              let checkIndices globals locals fname matrix indexList =
	               List.iter
	      	 	(fun exp ->
	      	        	let v = string_of_expr globals locals fname exp in
	      	                if v.vartype <> "int" then
	      	                raise (Failure ("Invalid index value: all indices "
	      	                ^ " must be int type. matrix: '" ^ matrix ^ "' ;"
	                          ^"function: '" ^ fname ^ "'"))) indexList
	      in
       
                let chkIn=checkIndices globals locals fname id indexList in
                blackhole chkIn;
                if NameMap.mem id locals then
                    let varId = NameMap.find id locals in
                    if varId.vartype <> "matrix" then
        	            raise (Failure ("Expecting matrix type but got '" ^ varId.vartype ^
        	                "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'")) 
        	        else { varname = id;vartype = "float" } 
        	  else if NameMap.mem id globals then
                    let varId = NameMap.find id globals in
                    if varId.vartype <> "matrix" then
        	            raise (Failure ("Expecting matrix type but got '" ^ varId.vartype ^
        	                "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'")) 
        	        else { varname = id;vartype = "float" } 
        	  else raise (Failure ("Undeclared identifier:" ^id^" found in function: '" ^ fname ^ "'"))      
       | MatrixRAccess(id, i1, i2, j1, j2) ->
       	          let ti1=string_of_expr globals locals fname i1 in
       	          let ti2=string_of_expr globals locals fname i2 in
       	          let tj1=string_of_expr globals locals fname j1 in
       	          let tj2=string_of_expr globals locals fname j2 in
       	          
       	         if ti1.vartype<>"int" || ti2.vartype<>"int" || tj1.vartype<>"int" || tj2.vartype<>"int" then raise (Failure("The range indices in " ^id^ " must by type int."))
                  else
                  if NameMap.mem id locals then
                          let varId = NameMap.find id locals in
                          if varId.vartype <> "matrix" then
              	            raise (Failure ("Expecting matrix type but got '" ^ varId.vartype ^
              	                "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'")) 
              	        else { varname = id;vartype = "matrix" } 
              	  else if NameMap.mem id globals then
                          let varId = NameMap.find id globals in
                          if varId.vartype <> "matrix" then
              	            raise (Failure ("Expecting matrix type but got '" ^ varId.vartype ^
              	                "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'")) 
              	        else { varname = id;vartype = "matrix" } 
        	  else raise (Failure ("Undeclared identifier:"^id^" found in function: '" ^ fname ^ "'"))
     | Newmat(row, col) ->
                     let varRow = string_of_expr globals locals fname row in
                     let varCol = string_of_expr globals locals fname col in
         	        if varRow.vartype <> varCol.vartype || varRow.vartype <> "int" then 
         	            raise (Failure ("Type mismatch in creating new matrix. Both row and column" 
         	                ^ " sizes must be of type int; in function: '" ^ fname ^ "'"))
                     else
    	            { varname = "newmat";vartype = "matrix" }
      | Call("mod", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 2 then 
                    raise (Failure ("'mod' function takes 2 parameters." 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param1 = List.nth actuals 1 in
                    let param2 = List.nth actuals 0 in
                        if param1.vartype = "int" && param2.vartype = "int" then
                            { varname = "mod" ; vartype = "int" }
                        else if param1.vartype = "float" && param2.vartype = "float" then
                            { varname = "mod" ; vartype = "float" }
                        else
                            raise (Failure ("Arguments passed to function 'mod' must be all of"
                            ^ " either int or float type,"
                            ^ " but here got '" ^ param1.vartype ^ "' and '" 
	                        ^ param2.vartype ^ "' called from function: '" ^ fname ^ "'"))      
      | Call("disp", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <>1 then 
                    raise (Failure ("Wrong number of arguments passed to function 'disp'" 
                        ^ ", called from function: '" ^ fname ^ "'"))

                
                else { varname = "disp" ; vartype = "void" }
     
      | Call("setFormat", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("Wrong number of arguments passed to function 'setFormat'" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                
                else 
                let param= List.hd actuals in
                if param.vartype<>"string" then 
                	
                	raise(Failure("The setFormat function only takes the following string parameters: 'short', 'shortE','long', 'longE'"))
                else if param.varname <>"short" && param.varname<>"shortE" && param.varname<>"long" && param.varname<>"longE" then
                	raise(Failure("The setFormat function only takes the following string parameters: 'short', 'shortE','long', 'longE'"))
                else { varname = "setFormat" ; vartype = "void" }                    
      | Call("tostring", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("'tostring' function only takes one parameter." 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype <> "matrix" then 
                            { varname = "tostring" ; vartype = "string" }
                        else
                            raise (Failure ("Argument passed to function 'tostring' cannot be " 
                                ^ "matrix type, called from function: '" ^ fname ^ "'"))
      | Call("load", actuals) ->
                  let actuals = List.fold_left 
                      (fun actuals actual ->
      	            let v = string_of_expr globals locals fdecl.fname actual in
      	                v :: actuals) [] actuals
                      in
                      if (List.length actuals) <> 1 then 
                          raise (Failure ("function 'load' only takes one string parameter " 
                              ^ ", called from function: '" ^ fname ^ "'"))
                      else
			let param= List.hd actuals in
                	if param.vartype<>"string" then                 	
                		raise(Failure("The 'load' function only takes the one parameter of type string"))
                	else                         
                   	 { varname = "load" ; vartype = "matrix" }
      | Call("save", actuals) ->
            		let actuals = List.fold_left 
	                    (fun actuals actual ->
	    	            let v = string_of_expr globals locals fdecl.fname actual in
	    	                v :: actuals) [] actuals
	                    in
	                    if (List.length actuals) <> 2 then 
	                        raise (Failure ("Function 'save takes 2 argument" 
	                            ^ ", called from function: '" ^ fname ^ "'"))
	                    else
	                        (*actuals list in reverse order*)
	                        let param2= (List.nth actuals 0) in
	                        let param1= (List.nth actuals 1) in
	                        if param1.vartype <>"matrix" then
	                        	raise(Failure(param1.vartype^"The first argument for the 'save' funnction must be a matrix"))
	                        else
	                        	if param2.vartype<>"string" then
	                        		raise(Failure("The second argument for the 'save' function must be a string"))
	                        	else
	                        		{ varname = "save" ; vartype = "void" }

      | Call("width", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("'width' function takes only one parameter." 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype = "matrix" then 
                            { varname = "width" ; vartype = "int" }
                        else
                            raise (Failure ("Argument passed to function 'width' must be of "
                            ^ "type matrix, called from function: '" ^ fname ^ "'"))

      | Call("inv", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("'inv' function takes only one parameter. " 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype = "matrix" then 
                            { varname = "inv" ; vartype = "matrix" }
                        else
                            raise (Failure ("Argument passed to function 'inv' must be of "
                            ^ "type matrix, called from function: '" ^ fname ^ "'"))
      | Call("height", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("'height' function takes only one parameter" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype = "matrix" then 
                            { varname = "height" ; vartype = "int" }
                        else
                            raise (Failure ("Argument passed to function 'height' must be of  "
                            ^ "type matrix, called from function: '" ^ fname ^ "'"))
      | Call("sqrt", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("function 'sqrt' takes just one parameter" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype = "float" || param.vartype = "int" then 
                            { varname = "sqrt" ; vartype = "float" }
                        else
                            raise (Failure ("Argument passed to function 'sqrt' can only be " 
                                ^ "either float or int type, called from function: '" ^ fname ^ "'"))
     | Call("abs", actuals) ->
                 let actuals = List.fold_left 
                     (fun actuals actual ->
     	            let v = string_of_expr globals locals fdecl.fname actual in
     	                v :: actuals) [] actuals
                     in
                     if (List.length actuals) <> 1 then 
                         raise (Failure ("function 'abs' takes just one parameter" 
                             ^ ", called from function: '" ^ fname ^ "'"))
                     else
                         let param = List.hd actuals in
                             if param.vartype = "float" || param.vartype = "int" then 
                                 { varname = "abs" ; vartype = "float" }
                             else
                                 raise (Failure ("Argument passed to function 'abs' can only be " 
                                ^ "either float or int type, called from function: '" ^ fname ^ "'"))
      | Call("exp", actuals) ->
                  let actuals = List.fold_left 
                      (fun actuals actual ->
      	            let v = string_of_expr globals locals fdecl.fname actual in
      	                v :: actuals) [] actuals
                      in
                      if (List.length actuals) <> 1 then 
                          raise (Failure ("function 'exp' takes just one parameter" 
                              ^ ", called from function: '" ^ fname ^ "'"))
                      else
                          let param = List.hd actuals in
                              if param.vartype = "float" || param.vartype = "int" then 
                                  { varname = "exp" ; vartype = "float" }
                              else
                                  raise (Failure ("Argument passed to function 'exp' can only be " 
                                ^ "either float or int type, called from function: '" ^ fname ^ "'"))
      | Call("log", actuals) ->
                  let actuals = List.fold_left 
                      (fun actuals actual ->
      	            let v = string_of_expr globals locals fdecl.fname actual in
      	                v :: actuals) [] actuals
                      in
                      if (List.length actuals) <> 1 then 
                          raise (Failure ("function 'log' takes just one parameter" 
                              ^ ", called from function: '" ^ fname ^ "'"))
                      else
                          let param = List.hd actuals in
                              if param.vartype = "float" || param.vartype = "int" then 
                                  { varname = "log" ; vartype = "float" }
                              else
                                  raise (Failure ("Argument passed to function 'log' can only be " 
                                ^ "either float or int type, called from function: '" ^ fname ^ "'"))
      | Call("log10", actuals) ->
                  let actuals = List.fold_left 
                      (fun actuals actual ->
      	            let v = string_of_expr globals locals fdecl.fname actual in
      	                v :: actuals) [] actuals
                      in
                      if (List.length actuals) <> 1 then 
                          raise (Failure ("function 'log' takes just one parameter" 
                              ^ ", called from function: '" ^ fname ^ "'"))
                      else
                          let param = List.hd actuals in
                              if param.vartype = "float" || param.vartype = "int" then 
                                  { varname = "log" ; vartype = "float" }
                              else
                                  raise (Failure ("Argument passed to function 'log' can only be " 
                                ^ "either float or int type, called from function: '" ^ fname ^ "'"))      
      | Call("sin", actuals) ->
                  let actuals = List.fold_left 
                      (fun actuals actual ->
      	            let v = string_of_expr globals locals fdecl.fname actual in
      	                v :: actuals) [] actuals
                      in
                      if (List.length actuals) <> 1 then 
                          raise (Failure ("function 'sin' takes just one parameter" 
                              ^ ", called from function: '" ^ fname ^ "'"))
                      else
                          let param = List.hd actuals in
                              if param.vartype = "float" || param.vartype = "int" then 
                                  { varname = "sin" ; vartype = "float" }
                              else
                                  raise (Failure ("Argument passed to function 'sin' can only be " 
                                ^ "either float or int type, called from function: '" ^ fname ^ "'"))
      | Call("cos", actuals) ->
                  let actuals = List.fold_left 
                      (fun actuals actual ->
      	            let v = string_of_expr globals locals fdecl.fname actual in
      	                v :: actuals) [] actuals
                      in
                      if (List.length actuals) <> 1 then 
                          raise (Failure ("function 'cos' takes just one parameter" 
                              ^ ", called from function: '" ^ fname ^ "'"))
                      else
                          let param = List.hd actuals in
                              if param.vartype = "float" || param.vartype = "int" then 
                                  { varname = "cos" ; vartype = "float" }
                              else
                                  raise (Failure ("Argument passed to function 'cos' can only be " 
                                ^ "either float or int type, called from function: '" ^ fname ^ "'"))
      | Call("tan", actuals) ->
                  let actuals = List.fold_left 
                      (fun actuals actual ->
      	            let v = string_of_expr globals locals fdecl.fname actual in
      	                v :: actuals) [] actuals
                      in
                      if (List.length actuals) <> 1 then 
                          raise (Failure ("function 'tan' takes just one parameter" 
                              ^ ", called from function: '" ^ fname ^ "'"))
                      else
                          let param = List.hd actuals in
                              if param.vartype = "float" || param.vartype = "int" then 
                                  { varname = "tan" ; vartype = "float" }
                              else
                                  raise (Failure ("Argument passed to function 'tan' can only be " 
                                ^ "either float or int type, called from function: '" ^ fname ^ "'"))
      | Call("asin", actuals) ->
                  let actuals = List.fold_left 
                      (fun actuals actual ->
      	            let v = string_of_expr globals locals fdecl.fname actual in
      	                v :: actuals) [] actuals
                      in
                      if (List.length actuals) <> 1 then 
                          raise (Failure ("function 'sqrt' takes just one parameter" 
                              ^ ", called from function: '" ^ fname ^ "'"))
                      else
                          let param = List.hd actuals in
                              if param.vartype = "float" || param.vartype = "int" then 
                                  { varname = "asin" ; vartype = "float" }
                              else
                                  raise (Failure ("Argument passed to function 'asin' can only be " 
                                ^ "either float or int type, called from function: '" ^ fname ^ "'"))
      | Call("acos", actuals) ->
                  let actuals = List.fold_left 
                      (fun actuals actual ->
      	            let v = string_of_expr globals locals fdecl.fname actual in
      	                v :: actuals) [] actuals
                      in
                      if (List.length actuals) <> 1 then 
                          raise (Failure ("function 'acos' takes just one parameter" 
                              ^ ", called from function: '" ^ fname ^ "'"))
                      else
                          let param = List.hd actuals in
                              if param.vartype = "float" || param.vartype = "int" then 
                                  { varname = "acos" ; vartype = "float" }
                              else
                                  raise (Failure ("Argument passed to function 'acos' can only be " 
                                ^ "either float or int type, called from function: '" ^ fname ^ "'"))
      | Call("atan", actuals) ->
                        let actuals = List.fold_left 
                            (fun actuals actual ->
            	            let v = string_of_expr globals locals fdecl.fname actual in
            	                v :: actuals) [] actuals
                            in
                            if (List.length actuals) <> 1 then 
                                raise (Failure ("function 'atan' takes just one parameter" 
                                    ^ ", called from function: '" ^ fname ^ "'"))
                            else
                                let param = List.hd actuals in
                                    if param.vartype = "float" || param.vartype = "int" then 
                                        { varname = "atan" ; vartype = "float" }
                                    else
                                        raise (Failure ("Argument passed to function 'atan' can only be " 
                                ^ "either float or int type, called from function: '" ^ fname ^ "'"))
      | Call("atan2", actuals) ->
                        let actuals = List.fold_left 
                            (fun actuals actual ->
            	            let v = string_of_expr globals locals fdecl.fname actual in
            	                v :: actuals) [] actuals
                            in
                            if (List.length actuals) <> 2 then 
                                raise (Failure ("function 'atan2' takes 2 parameter" 
                                    ^ ", called from function: '" ^ fname ^ "'"))
                            else
                                let param = List.hd actuals in
                                    if param.vartype = "float" || param.vartype = "int" then 
                                        { varname = "atan" ; vartype = "float" }
                                    else
                                        raise (Failure ("Argument passed to function 'atan' can only be " 
                                ^ "either float or int type, called from function: '" ^ fname ^ "'"))   
      | Call(f, actuals) ->
            let fdecl =
                if f = "main" then
                    raise (Failure ("main function cannot be called by " ^
	                                         "function: '" ^ fname ^ "'"))
                else
	            try NameMap.find f func_decls
	                with Not_found -> raise (Failure ("Undefined function: '" 
	                    ^ f ^ "' in function call: '" ^ fname ^ "'"))
	            in
	            if f <> fname then
    	            let actuals = List.fold_left 
    	                (fun actuals actual ->
    		            let v = string_of_expr globals locals fdecl.fname actual in
    		                v :: actuals) [] actuals
    	                in
                            string_of_fdecl globals fdecl (List.rev actuals);
                            { varname = fdecl.retname ; vartype = fdecl.rettype }
                else
                    let actuals = List.fold_left 
    	                (fun actuals actual ->
    		            let v = string_of_expr globals locals fdecl.fname actual in
    		                v :: actuals) [] actuals
    	                in
                (try List.iter2 (fun formal actual -> 
                    if formal.vartype <> actual.vartype then
                        raise (Failure ("Type mismatch in argument passing between: '" 
                        ^ formal.varname ^ "', type: " ^ formal.vartype ^ " and '" 
                        ^ actual.varname ^ "', type: " ^ actual.vartype 
                        ^ " in function: '" ^ fdecl.fname ^ "'")))
                       fdecl.formals (List.rev actuals)
                (* Check invalid number of arguments *)    
                with Invalid_argument(_) ->
                    raise (Failure ("Wrong number of arguments passed to function: '" 
                        ^ fdecl.fname ^ "' from function: '" ^ fname ^ "'")));
                { varname = fdecl.retname ; vartype = fdecl.rettype }
    in
    
    let check_for_expr v l globals locals fname =
        if v.vartype <> "int" or v.vartype <> "float" then
            let varList = List.map (string_of_expr globals locals fname) l
	        in
    	    List.iter (fun var -> if var.vartype <> v.vartype then  
    	            raise (Failure ("For loop type mismatch between '" ^ v.varname 
    	            ^ "', type: " ^ v.vartype ^ " and '" ^ var.varname 
    	            ^ "', type: " ^ var.vartype ^ " in function: '" ^ fname ^ "'"))) varList;
	    else
	        raise (Failure ("Only int/float is allowed in the for loop. Id '" 
	            ^ v.varname ^ "', type: " ^ v.vartype ^ " found in function: '" ^ fname ^ "'"))
    in	        
    
    let rec string_of_for_expr globals locals fname = function
        Assigna(v, e1, e2) -> 
        if NameMap.mem v locals then
            let var = NameMap.find v locals in
            check_for_expr var [e1;e2;] globals locals fname
        else if NameMap.mem v globals then
	        let var = NameMap.find v globals in
	        check_for_expr var [e1;e2;] globals locals fname
	    else raise (Failure ("Undeclared identifier in for loop: '" ^ v 
	        ^ "' in function: '" ^ fname ^ "'"))
    |   Assignb(v, e1, e2, e3) ->
        if NameMap.mem v locals then
            let var = NameMap.find v locals in
            check_for_expr var [e1;e2;e3;] globals locals fname
	    else if NameMap.mem v globals then
	        let var = NameMap.find v globals in
	        check_for_expr var [e1;e2;e3] globals locals fname
        else raise (Failure ("Undeclared identifier in for loop: '" ^ v
            ^ "' in function: '" ^ fname ^ "'"))
    in
    let checkDimension matrix fname l = 
        let rec loop l = 
            if (List.length l) = 1 then 
                List.length (List.hd l)
            else
                let c1 = List.length (List.hd l)
                and c2 = loop (List.tl l) in
                if (c1 <> c2) then
                    raise (Failure ("Invalid matrix initialization: number of columns"
                    ^ " in all rows must be same. matrix: '" ^ matrix ^ "' ;"
                    ^"function: '" ^ fname ^ "'"))
                else c1
            in loop l
    in
    let checkMatrixValues globals locals matrix fname l = 
        let flatL = List.concat l in
            List.iter
            (fun exp ->
            let v = string_of_expr globals locals fname exp in
                if v.vartype <> "float" && v.vartype<> "int" then
                    raise (Failure ("Invalid matrix value: all values passed to "
                        ^ " matrix must be float type. matrix: '" ^ matrix ^ "' ;"
                    ^"function: '" ^ fname ^ "'"))) flatL
    in
    let rec string_of_stmt globals locals fname = function
        Block(stmts) ->
            List.iter (string_of_stmt globals locals fname) stmts
      | Expr(expr) -> let e = string_of_expr globals locals fname expr in blackhole e
      | If(e, s1, s2) ->  let v = string_of_expr globals locals fname e in 
            if v.vartype <> "boolean" then
                raise (Failure ("if expression must be evaluated to boolean, "
                    ^ "but here found with '" ^ v.vartype ^ "' ;"
                    ^ "function: '" ^ fname ^ "'"))
            else
                string_of_stmt globals locals fname s1;
                string_of_stmt globals locals fname s2;
      | Ifelseif(e, s1, eif_list, s2) -> 
            let v = string_of_expr globals locals fname e in 
            if v.vartype <> "boolean" then
                raise (Failure ("if expression must be evaluated to boolean, "
                    ^ "but here found with '" ^ v.vartype ^ "' ;"
                    ^ "function: '" ^ fname ^ "'"))
            else
                let rec string_of_elseif eif = 
                    let en = string_of_expr globals locals fname eif.elseif_expr in
                        if en.vartype <> "boolean" then
                            raise (Failure ("if expression must be evaluated to boolean, "
                                ^ "but here found with '" ^ v.vartype ^ "' ;"
                                ^ "function: '" ^ fname ^ "'"))
                        else
                            string_of_stmt globals locals fname eif.elseif_stmt;
                in
                    string_of_stmt globals locals fname s1;
                    List.iter string_of_elseif eif_list;
                    string_of_stmt globals locals fname s2;
      | For(e, s) ->loop_flag :=  true;
                string_of_for_expr globals locals fname e;
                string_of_stmt globals locals fname s;
                loop_flag := false;
      | While(e, s) -> loop_flag:=true;
            let v = string_of_expr globals locals fname e 
            in 
            if v.vartype <> "boolean" then
                raise (Failure ("while expression must be evaluated to boolean, "
                    ^ "but here found with '" ^ v.vartype ^ "' ;"
                    ^ "function: '" ^ fname ^ "'"))
            else
                string_of_stmt globals locals fname s;
            loop_flag:=false;
      | MatrixInit(id, valuesList) ->
            if NameMap.mem id locals then
                let varId = NameMap.find id locals in
                if varId.vartype <> "matrix" then
    	            raise (Failure ("Expecting matrix type but got '" ^ varId.vartype ^
    	                "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'")) 
    	        else 
    	            let col = checkDimension id fname valuesList in
    	                blackhole col;
    	                checkMatrixValues globals locals id fname valuesList
    	    else if NameMap.mem id globals then
                let varId = NameMap.find id globals in
                if varId.vartype <> "matrix" then
    	            raise (Failure ("Expecting matrix type but got '" ^ varId.vartype ^
    	                "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'")) 
    	        else
    	            let col = checkDimension id fname valuesList in
    	                blackhole col;
    	                checkMatrixValues globals locals id fname valuesList
    	    else raise (Failure ("Undeclared identifier:"^id^" found in function: '" ^ fname ^ "'"))
       | Break -> if !loop_flag = false then
       			raise (Failure ("Misplaced break statement is found outside of a loop"))
      		  else
      		  ()
      		  
       | Continue ->if !loop_flag = false then
       			raise (Failure ("Misplaced continue statement is found outside of a loop"))
      		  else
      		  ()
    in
    let locals = 
        try List.fold_left2
            (fun locals formal actual -> 
                if formal.vartype = actual.vartype then
                    if NameMap.mem formal.varname locals then
                        raise (Failure ("Formal variable: '" ^ formal.varname ^ 
                            "' has already been defined in the function: '" ^ 
                            fdecl.fname ^ "' parameter"))
                    else NameMap.add formal.varname formal locals
                else raise (Failure ("Type mismatch in argument passing between: '" 
    	                ^ formal.varname ^ "', type: " ^ formal.vartype ^ " and '" 
    	                ^ actual.varname ^ "', type: " ^ actual.vartype 
    	                ^ " in function: '" ^ fdecl.fname ^ "'")))
	            NameMap.empty fdecl.formals actuals
	    (* Check invalid number of arguments *)    
        with Invalid_argument(_) ->
	        raise (Failure ("Wrong number of arguments passed to function: '" 
	            ^ fdecl.fname ^ "'"))    
        in      
        let locals = List.fold_left
	        (fun locals local -> 
                if NameMap.mem local.varname locals then
                        raise (Failure ("Local variable: '" ^ local.varname ^ 
                            "' has already been defined in the function: '" ^ 
                            fdecl.fname ^ "'"))
                else NameMap.add local.varname local locals) locals fdecl.locals
            in  
            if fdecl.retname <> "" then
                if NameMap.mem fdecl.retname locals then 
                    raise (Failure ("Return variable: '" ^ fdecl.retname ^ 
                        "' has already been defined in the function: '" ^ 
                        fdecl.fname ^ "'"))
                else
                    let ret = { varname = fdecl.retname; vartype = fdecl.rettype } in
                    let locals = NameMap.add ret.varname ret locals in
                    List.iter (string_of_stmt globals locals fdecl.fname) fdecl.body
            else
                List.iter (string_of_stmt globals locals fdecl.fname) fdecl.body
    in
    let globals = List.fold_left
        (fun globals vdecl -> 
            if NameMap.mem vdecl.varname globals then
                raise (Failure ("Global variable : '" ^ vdecl.varname ^ 
                            "' has already been defined."))
            else
                NameMap.add vdecl.varname vdecl globals) NameMap.empty vars
    in 
        try
            string_of_fdecl globals (NameMap.find "main" func_decls) []
        with Not_found -> 
            raise (Failure ("There is no main() function"))        

